﻿/* ファイルリネーマオルカ (FileInfoText)
 * 
 * COPYLEFT (C') CharMY 2013 Yanagihara Masami
 * 
 * 
 * This program is free software: you can redistribute it and/or modify   *
 * it under the terms of the GNU General Public License as published by   *
 * the Free Software Foundation, either version 3 of the License, or      *
 * (at your option) any later version.                                    *
 *                                                                        *
 * This program is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 * 
 * 
 */

#if DEBUG
//#define NOMOVE
#endif

using System;
using System.IO;
using System.Linq;
using CharMY.ComponentModel;
using CharMY.IO;
using CharMY.Text;
using Shell32;

namespace CharMY.Orca
{
    internal class FileInfoText : CharMYCore.NonElement
    {
        /// <summary>
        /// The text equals to a part of file path.
        /// </summary>
        public string Text
        {
            get { return _text; }
            set
            {
                _text = (value ?? _text)
                    .ToFileSystemValidChars(drive: _dir_level < 0, dir: _dir_level != 0);
            }
        }
        public string Original
        {
            get
            {
                var result = _info.FullName
                    .ReplaceByRegexWithoutTags(@"[\\/]+", Chars.FilePathDirectorySeparator2);
                result = EditDirectories(_info, result, -1, _dir_level);
                result = EditExtension(_info, result, true, _ext);
                return result;
            }
        }
        public FileInfo File { get { return _info; } }
        private FileInfo _info;
        private string _text;
        private bool _ext = true;
        private int _dir_level = 0;
        /// <summary>
        /// <para>level >= 0; The text contains a series of (level - 1) names of directories starting with the parent directory (if exists) but the drive name.</para>
        /// <para>level < 0; The text does NOT contain a series of (-level) names of directories starting with the drive name.</para>
        /// <para>This definition means as following ...</para>
        /// <para> 0; The text contains only a name of the file. (without any directories' path)</para>
        /// <para>-1; The text equals to the full path of the file. (with the drive name, which only this level contains.)</para>
        /// <para>-2; The text contains all parent directories' names but the drive.</para>
        /// <para> 1; The tex  equals to a set of the name and the parent directory of the file. (if exists; the directory is not the drive)</para>
        /// </summary>
        public int DirectoryLevel
        {
            get { return _dir_level; }
            set
            {
                _text = EditDirectories(_info, _text, _dir_level, value);
                _dir_level = value;
            }
        }
        public bool WithExtension
        {
            get { return _ext; }
            set
            {
                _text = EditExtension(_info, _text, _ext, value);
                _ext = value;
            }
        }

        public FileInfo Target()
        {
            var path = _text;
            path = EditDirectories(_info, path, _dir_level, -1, _dir_level);
            path = EditExtension(_info, path, _ext, true);
            Assert.WriteLine(_text + " > " + path);
            var result = path.ToFile();
            return result;
        }

        private static string EditExtension(FileInfo info, string path, bool argWithExt, bool resultWithExt)
        {
            Assert.NotNull(info);
            Assert.NotNull(path);
            if (argWithExt && !resultWithExt)
            {
                //delete ext
                if (path.EndsWith(info.Extension))
                    path = path.Substring(0, path.Length - info.Extension.Length);
            }
            else if (!argWithExt && resultWithExt)
            {
                //add ext
                path += info.Extension;
            }
            return path;
        }
        private static string EditDirectories(FileInfo info, string path, int argLevel, int resultLevel, int pathLevel = 0)
        {
            Assert.NotNull(info);
            Assert.NotNull(path);
            if (argLevel == resultLevel)
                return path;

            var divided = path
                .SplitByRegex(@"[\\/:]+")
                .ToArray();

            if (argLevel < 0)
            #region
            {
                if (resultLevel < argLevel)
                {
                    //delete sub
                    path = EditDirDeleteSub(divided, argLevel - resultLevel);
                }
                else if (resultLevel < 0)
                {
                    //add parent but cut>>>
                    var dirs = info.Directory.GetParentDirectories().ToArray();
                    path = EditDirAddParent(divided, argLevel, dirs, resultLevel - argLevel, -(resultLevel + 1));
                }
                else
                {
                    var dirs = info.Directory.GetParentDirectories().ToArray();
                    if (resultLevel < dirs.Length + argLevel + 1)
                    {
                        ///delete sub but remain enum<<<
                        path = EditDirDeleteSub(divided, (dirs.Length + argLevel + 1) - resultLevel, resultLevel);
                    }
                    else
                    {
                        //add parent but drive in enum<<<
                        path = EditDirAddParent(divided, argLevel, dirs, resultLevel - (dirs.Length + argLevel + 1));
                    }
                }
            }
            #endregion
            else
            #region
            {
                if (argLevel <= resultLevel)
                {
                    //add parent but drive in enum<<<
                    var dirs = info.Directory.GetParentDirectories().ToArray();
                    path = EditDirAddParent(divided, argLevel, dirs, resultLevel - argLevel);
                }
                else if (0 <= resultLevel)
                {
                    //delete sub but remain enum<<<
                    path = EditDirDeleteSub(divided, argLevel - resultLevel, resultLevel);
                }
                else
                {
                    var dirs = info.Directory.GetParentDirectories().ToArray();
                    if (argLevel <= dirs.Length + resultLevel + 1)
                    {
                        //add parent but cut>>>
                        path = EditDirAddParent(divided, argLevel, dirs, (dirs.Length + resultLevel + 1) - argLevel, -(resultLevel + 1));
                    }
                    else
                    {
                        //delete sub
                        path = EditDirDeleteSub(divided, argLevel - (dirs.Length + resultLevel + 1));
                    }
                }
            }
            #endregion

            if (resultLevel == 0)
            {
                path = path.ToFileSystemValidChars(dir: false);
            }
            else if (resultLevel == -1 && !path.Contains(":"))
            {
                path = info.Directory.Root.Name.Replace("\\", "/") + path;
            }

            return path;
        }
        private static string EditDirDeleteSub(string[] divided, int level, int remain = 0)
        {
            Assert.True(level >= 0);
            Assert.True(remain >= 0);
            var result = divided
                .SkipWhile((_, i) => (i < level) && (i < divided.Length - 1 - remain))
                .ToSingleString(Chars.FilePathDirectorySeparator2);
            return result;
        }
        private static string EditDirAddParent(string[] divided, int argLevel, DirectoryInfo[] dirs, int level, int cutMust = 1)
        {
            Assert.True(level >= 0);
            Assert.True(cutMust >= 0);
            var result = divided
                .Reverse()
                .Concat(dirs
                    .Skip(argLevel >= 0 ? argLevel : (divided.Length - 1))
                    .Reverse()
                    .Skip(cutMust)
                    .Reverse()
                    .Take(level)
                    .Select(dir => dir.Name.Replace("\\", ""))
                )
                .Reverse()
                .ToSingleString(Chars.FilePathDirectorySeparator2);
            return result;
        }

        public bool? FileRename(bool deleteEmptyDirectorie = false)
        {
            if (!_info.Exists)
                return null;

            var target_core = this.Target();
            if (target_core == null)
                return null;

            if (_info.PathEquals(target_core))
                return true;

            bool result = true;
            try
            {
                var target = target_core;
                for (int i = 2 ; target.Exists ; ++i)
                {
                    target = target_core.ChangeName((name, ext) => name + i + ext);
                    result = false;

                    if (_info.PathEquals(target))
                        return result;
                }


                Assert.WriteLine(_info.ToInformation());
                Assert.WriteLine("move to");
                Assert.WriteLine(target.ToInformation());
                Assert.WriteLine();

                var dir = _info.Directory;
#if !NOMOVE
                target.Directory.CreateDirectory();
                _info.MoveTo(target.FullName);
                _info = new FileInfo(target.FullName);

                Assert.True(_info.PathEquals(target));
                Assert.True(_info.Exists);
#endif
                if (deleteEmptyDirectorie && (dir.GetFiles().Length + dir.GetDirectories().Length) == 0)
                    dir.Delete(false);
            }
            catch (Exception e) { Console.WriteLine(e.Message); }
            return result;
        }
        public bool? FileCopy()
        {
            if (!_info.Exists)
                return null;

            var target = this.Target();
            if (target == null)
                return null;

            if (_info.PathEquals(target))
                return true;

            if (target.Exists)
                return false;

            try
            {
                Assert.WriteLine(_info.ToInformation());
                Assert.WriteLine("copy to");
                Assert.WriteLine(target.ToInformation());
                Assert.WriteLine();
#if !NOMOVE
                target.Directory.CreateDirectory();
                _info.CopyTo(target.FullName);
                _info = new FileInfo(target.FullName);

                Assert.True(_info.PathEquals(target));
                Assert.True(_info.Exists);
#endif
            }
            catch (Exception e) { Console.WriteLine(e.Message); }
            return true;
        }

        public FileInfoText(FileInfo info)
        {
            Checker.ArgumentNotNull(info);
            _info = info;
            _text = _info.Name;
        }

        public string GetUpdateDate()
        {
            var result = _info.LastAccessTime.ToShortDateString();
            return result;
        }
        public string GetUpdateTime()
        {
            var result = _info.LastWriteTime.ToShortTimeString();
            return result;
        }
        public string GetSize()
        {
            var len = _info.Length;
            if (len < 0x200)
            {
                var result = len.ToString() + "byte";
                return result;
            }
            else if (len < 0x8000)
            {
                var result = (len / (float)0x400).ToString("F1") + "KB";
                return result;
            }
            //                ..  ..
            else if (len < 0x2000000)
            {
                //                           .  ..
                var result = (len / (float)0x10000).ToString("F1") + "MB";
                return result;
            }
            //                 ..  ..
            else if (len < 0x80000000)
            {
                //                            ..  ..
                var result = (len / (float)0x4000000).ToString("F1") + "GB";
                return result;
            }
            //                ..  ..  ..
            else if (len < 0x20000000000)
            {
                //                           .  ..  ..
                var result = (len / (float)0x100000000).ToString("F1") + "TB";
                return result;
            }
            else
            {
                //                            ..  ..  ..
                var result = (len / (float)0x40000000000).ToString("F1") + "PB";
                return result;
            }
        }

        static readonly Shell __shell = new Shell();
        public enum FileTagIndex
        {
            LineName = 0,
            FileSize = 1,
            FileKind = 2,
            Update = 3,
            Create = 4,
            Access = 5,
            Attribute = 6,

            FileType = 9,
            Owner = 10,
            FileType2 = 11,

            AlbumArtist = 13,
            Album = 14,
            Year = 15,
            Genre = 16,
            Commander = 17,

            Evaluation = 19,
            Artist = 20,
            Title = 21,

            Comment = 24,
            No = 26,
            TuneLength = 27,
            TuneWidth = 28,
            Protect = 29,
        }
        public string GetMP3Tag(FileTagIndex index)
        {
            var dir = __shell.NameSpace(_info.Directory.FullName);
            var item = dir.ParseName(_info.Name);

            try
            {
                var value = dir.GetDetailsOf(item, (int)index);
                return value ?? "";
            }
            catch (Exception) { }

            return "";
        }

        public override string ToString() { return this.Text; }
    }
}
