﻿/* ファイルリネーマオルカ
 * 
 * COPYLEFT (C') CharMY 2013 Yanagihara Masami
 * 
 * 
 * This program is free software: you can redistribute it and/or modify   *
 * it under the terms of the GNU General Public License as published by   *
 * the Free Software Foundation, either version 3 of the License, or      *
 * (at your option) any later version.                                    *
 *                                                                        *
 * This program is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 * 
 * 
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using CharMY.IO;
using CharMY.Resources;
using CharMY.Text;
using CharMY.Windows.Forms;

namespace CharMY.Orca
{
    using Menu = MenuStripItemAttribute;
    using Tool = ToolStripItemAttribute;
    internal class FileRenamerMenuSet : ColoringTextMenuSet
    {
        private readonly ToolStripMenuItem _keep_line_strip_item = new ToolStripMenuItem();
        private readonly ValueWatcher<bool> _text_mode = new ValueWatcher<bool>();
        private void SetLinerEdit()
        {
            Assert.NotNull(_line_keeper);

            var keep_line = new ValueWatcher<bool>();
            keep_line.SetSyncChecked(_keep_line_strip_item);
            keep_line.ValueChanged += () => _line_keeper.KeepLine = keep_line.Value;

            _text_mode.ValueChanged += () =>
            {
                if (_text_mode.Value)
                {
                    _keep_line_strip_item.Enabled = true;
                    keep_line.Value = false;
                }
                else
                {
                    _keep_line_strip_item.Enabled = false;
                    keep_line.Value = true;
                }
            };

            keep_line.Value = false;
            _line_keeper.KeepLine = false;
            _text_mode.Value = true;

            _dir_level.MaxValue = 15;
            _dir_level.MinValue = -16;
            _dir_level.Value = 0;
            _dir_level.ValueChanged += this.UpdateFileState;
            _ext_mode.ValueChanged += this.UpdateFileState;
            _ext_mode.Value = true;

            this.Content.LineSwap = (i, j) =>
            {
                if (_file_list != null)
                {
                    if (_file_list.Length <= i || _file_list.Length <= j)
                        return false;

                    _file_list.Swap(i, j);
                }
                return true;
            };
        }

        protected override bool StrictEachLineSearch { get { return true; } }
        protected override KeyValuePair<string, Func<int, string>>[] MakeLineTags()
        {
            // 置換タグ作成
            var type = TypeInference<int>.Factory;
            var result = base.MakeLineTags()
                .ConcatItem(
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetUpdateDate()).Cache().WithKey("UpdateDate"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetUpdateTime()).Cache().WithKey("UpdateTime"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetSize()).Cache().WithKey("Size"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Title)).Cache().WithKey("Title"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Album)).Cache().WithKey("Album"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Artist)).Cache().WithKey("Artist"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Genre)).Cache().WithKey("Genre"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.No)).Cache().WithKey("Number"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Commander)).Cache().WithKey("Commander"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Comment)).Cache().WithKey("Comment"),
                    type.Func(i => _file_list == null ? "" : _file_list[i % _file_list.Length].GetMP3Tag(FileInfoText.FileTagIndex.Evaluation)).Cache().WithKey("Evaluation")
                )
                .ToArray();
            return result;
        }

        private readonly DragDropFiles _drop_files = new DragDropFiles();
        private readonly FileTextCorrector _line_keeper;
        protected sealed override TextEditor MakeCoreEditor() { return new FileTextCorrector(this, base.MakeCoreEditor()); }
        #region
        private class FileTextCorrector : TextLineKeeper
        {
            private readonly FileRenamerMenuSet _owner;
            public FileTextCorrector(FileRenamerMenuSet owner, TextEditor @base) : base(@base) { _owner = owner; }
            protected override void CheckValuesImpl(ref StringChangedRanges changedText, ref IndexRange selection)
            {
                if (changedText == null)
                    return;

                if (!this.KeepLine)
                    return;

                base.CheckValuesImpl(ref changedText, ref selection);

                var corrected = changedText.Correct(_owner.Correct);
                if (object.ReferenceEquals(corrected, changedText))
                    return;

                this.CorrectSelection(corrected, ref selection);
                changedText = corrected;
            }
        }
        private string Correct(StringChangedRanges.ChangedRange range)
        {
            // ファイルパスに使用できない文字を置換
            var result = Chars.ToFileSystemValidChars(range.TextInResult,
                drive: _dir_level.Value == -1,
                dir: _dir_level.Value != 0,
                replacement: "_");

            if (range.TextInResult == result)
                return null;

            return result;
        }
        #endregion

        private FileInfoText[] _file_list = null;
        public void SetFiles(FileInfo[] files)
        {
            string[] lines = null;
            if (files == null || files.Length == 0)
            {
                _file_list = null;
                this.Content.ListItems = i => (i + 1).ToString();
                Console.WriteLine("It is Text Editor Mode");
            }
            else
            {
                _line_keeper.KeepLine = false;
                if (_file_list != null)
                    lines = this.Editor.Text.SplitToLines();

                var add = files
                    .Where()
                    .Select(info => new FileInfoText(info));

                _file_list = (_file_list ?? _file_list.ToEmpty())
                    .Concat(add)
                    .ToArray();
                this.Content.ListItems = i => _file_list[i % _file_list.Length].Original;
                Console.WriteLine("It is File Renamer Mode");
            }

            this.UpdateFileState(lines);
            _text_mode.Value = _file_list == null;
            _line_keeper.KeepLine = _file_list != null;
            this.Content.ResetList();
        }
        private readonly ValueWatcher<bool> _ext_mode = new ValueWatcher<bool>();
        private readonly ValueTextBinderInt _dir_level = new ValueTextBinderInt();
        private void UpdateFileState() { this.UpdateFileState(this.Editor.Text.SplitToLines()); }
        private void UpdateFileState(string[] currentLines)
        {
            if (_file_list == null)
                return;

            if (currentLines != null)
                foreach (var pair in currentLines.ZipKey(_file_list))
                    pair.Key.Text = pair.Value;

            foreach (var file in _file_list)
            {
                file.WithExtension = _ext_mode.Value;
                file.DirectoryLevel = _dir_level.Value;
            }
            base.ImportText(_file_list.Select(file => file.Text).ToSingleStringLines());
            this.Content.ResetList();
        }

        #region メニューGUIの自動生成

        public override void Open()
        {
            if (_file_list == null)
            {
                base.Open();
            }
            else
            {
                this.AddFilesForRename();
            }
        }
        public override void Save()
        {
            if (_file_list == null)
            {
                base.Save();
            }
            else
            {
                this.Rename();
            }
        }

        private const int indexOrca = 203;

        [Tool(1, "ADD", indexOrca, "Renamer", BaseStrip = StripPlace.ToolTop)]
        [Menu(1 | indexSeparator, indexFile)]
        public void AddFilesForRename()
        {
            var files = Files.Any().GetFiles(title: "SELECT FILES FOR REMAMING.");
            if (files != null && files.Length > 0)
                this.SetFiles(files);
        }
        [Tool(2, "RENAME", indexOrca, null)]
        [Menu(2 | indexSeparator, indexFile)]
        public void Rename()
        {
            if (_file_list == null)
            {
                this.Version();
                return;
            }

            this.UpdateFileState();

            var cands = _file_list
                .GroupBy(file => file.File.FullName.ReplaceByRegexWithoutTags(@"[\\/]+", "/"))
                .SelectMany(group => group
                    .Skip(1)
                    .ZipKey(_ => -1)
                    .ConcatItem(group.First().Cache().WithKey(0))
                )
                .ToQueue();

            var cands_copy = cands.Where(c => c.Key < 0).ToQueue();
            cands = cands.Where(c => c.Key >= 0).ToQueue();

            while (cands_copy.Count > 0)
            {
                var cand = cands_copy.Dequeue();
                var result = cand.Value.FileCopy();
                if (result == null || result.Value || cand.Key == cands_copy.Count)
                    continue;

                //retry
                cands_copy.Enqueue(cand.Value.Cache().WithKey(cands_copy.Count));
            }

            while (cands.Count > 0)
            {
                var cand = cands.Dequeue();
                var result = cands_copy.Any(cc => cc.Value.File.PathEquals(cand.Value.File)) ?
                    cand.Value.FileCopy() :
                    cand.Value.FileRename(this.DeleteEmptyDirectory);
                if (result == null || result.Value || cand.Key == cands.Count)
                    continue;

                //retry
                cands.Enqueue(cand.Value.Cache().WithKey(cands.Count));
            }

            Assert.True(_file_list.All(file => file.File.Exists));

            this.Content.ResetList();
            this.UpdateFileState();
        }

        [Tool(3, "CLEAR", indexOrca, null)]
        [Menu(3 | indexSeparator, indexFile)]
        public void ClearFilesForRename() { this.SetFiles(null); }

        [Menu(4 | (indexSeparator * 4), indexEdit)]
        public ToolStripItem KeepLine { get { return _keep_line_strip_item; } }
        [Tool(5, "EXT", indexOrca, null)]
        [Menu(5 | (indexSeparator * 4), indexEdit)]
        public ValueWatcher<bool> ShowExtension { get { return _ext_mode; } }
        [Tool(6, "DIR", indexOrca, null)]
        [Menu(6 | (indexSeparator * 4), indexEdit)]
        public void DirectoryLevel()
        {
            _dir_level.Value = WindowForms.ValueInput(
                _dir_level.MinValue,
                _dir_level.MaxValue, 0)
                .ShowDialogSelectOrDefault(
                    _dir_level.Value,
                    _dir_level.Value,
                    this.PageInfo.Window,
                    "DirectoryLevel"
                ).ToInt();
        }
        [Tool(7, indexOrca)]
        [Menu(7 | (indexSeparator * 4), indexEdit)]
        public ValueTextBinderInt DirectoryLevelBox { get { return _dir_level; } }
        [Menu(8 | (indexSeparator * 4), indexEdit)]
        public bool DeleteEmptyDirectory { get; set; }

        [Menu(1, 15 | (indexSeparator * 3), "FileTags", indexEdit)]
        public void InsertFileDate() { this.InsertFileCore(f => f.GetUpdateDate()); }
        [Menu(2, 15 | (indexSeparator * 3), indexEdit)]
        public void InsertFileTime() { this.InsertFileCore(f => f.GetUpdateTime()); }
        [Menu(3, 15 | (indexSeparator * 3), indexEdit)]
        public void InsertFileSize() { this.InsertFileCore(f => f.GetSize()); }
        [Menu(4, 15 | (indexSeparator * 3), indexEdit)]
        public void InsertMP3Title() { this.InsertFileCore(f => f.GetMP3Tag(FileInfoText.FileTagIndex.Title)); }
        private void InsertFileCore(Func<FileInfoText, string> toTag)
        {
            this.Content.Paste(_file_list == null ? " - no file exists - " :
                toTag(_file_list[this.Editor.Text.LineIndex(this.Editor.Selection.Start) % _file_list.Length]));
        }

        #endregion

        public FileRenamerMenuSet(PageInfo page)
            : base(page)
        {
            this.DeleteEmptyDirectory = true;
            _line_keeper = base.Editor.CompositeBase as FileTextCorrector;
            this.SetLinerEdit();

            var keep = false;
            this.Content.ColoringModeChanging += () =>
            {
                keep = _line_keeper.KeepLine;
                _line_keeper.KeepLine = false;
            };
            this.Content.ColoringModeChanged += () => _line_keeper.KeepLine = keep;

            _drop_files.SetDragDrop(this.Content);
            _drop_files.DragDroped = this.SetFiles;

            var initial_text = "Drag & drop files to here\nfor file renaming mode.";
            this.ImportText(initial_text);
            page.Loaded += () =>
            {
                page.Window.Icon = Logo.Icon;
                if (this.Editor.TextLength == 0)
                {
                    this.ImportText(initial_text);
                }
            };
        }

        protected override Image VersionImageImpl() { return Logo.Image; }
        protected override string VersionNameImpl() { return "公開されたコードによるビルド"; }
        public override string ToString() { return "Orca"; }
    }
}
